#!/usr/bin/env python3

import hashlib

from biom import load_table, Table
from biom.util import biom_open
from Bio import SeqIO
from Bio.SeqIO import FastaIO


def hash_function(seq):
    return hashlib.md5(seq.encode("utf-8")).hexdigest()


def hash_otu_table(unhashed_otu_table, output_file):
    table = load_table(unhashed_otu_table)
    df = table.to_dataframe(dense=True)
    seq_ids = list(map(hash_function, df.index))
    df.index = seq_ids
    new_table = Table(df.values, list(df.index), list(df.columns))
    with biom_open(output_file, "w") as fid:
        new_table.to_hdf5(fid, "Constructed by micone in dada2 pipeline")
    return seq_ids


def hash_rep_seqs(unhashed_rep_seqs, output_file):
    seqs = list(SeqIO.parse(unhashed_rep_seqs, "fasta"))
    seq_ids = []
    for seq in seqs:
        seq.id = hash_function(str(seq.seq))
        seq_ids.append(seq.id)
        seq.description = ""
        seq.name = ""
    with open(output_file, "w") as fid:
        fasta_writer = FastaIO.FastaWriter(fid, wrap=None)
        fasta_writer.write_file(seqs)
    return seq_ids


def hashing(unhashed_otu_table, unhashed_rep_seqs):
    otu_table_ids = hash_otu_table(unhashed_otu_table, "otu_table.biom")
    rep_seq_ids = hash_rep_seqs(unhashed_rep_seqs, "rep_seqs.fasta")
    assert otu_table_ids == rep_seq_ids


if __name__ == "__main__":
    UNHASHED_OTU_TABLE = "${unhashed_otu_table}"
    UNHASHED_REP_SEQS = "${unhashed_rep_seqs}"
    hashing(UNHASHED_OTU_TABLE, UNHASHED_REP_SEQS)



#!/usr/bin/env python3

from micone import Network, NetworkGroup


def main(
    base_name: str,
    corr_file: str,
    meta_file: str,
    cmeta_file: str,
    obsmeta_file: str,
    pvalue_file: str,
    children_file: str,
) -> None:
    network = Network.load_data(
        interaction_file=corr_file,
        meta_file=meta_file,
        cmeta_file=cmeta_file,
        obsmeta_file=obsmeta_file,
        pvalue_file=pvalue_file,
        children_file=children_file,
        interaction_threshold=0.2,
        pvalue_threshold=0.05,
    )
    network_group = NetworkGroup([network])
    network_group.write(base_name + "_network.json")
    network_group.write(
        base_name + "_thres_network.json", pvalue_filter=False, interaction_filter=True
    )


if __name__ == "__main__":
    BASE_NAME = "${level}"
    CORR_FILE = "${corr_file}"
    META_FILE = "${metadata}"
    CMETA_FILE = "$cmetadata"
    OBSMETA_FILE = "$obsdata_file"
    PVALUE_FILE = "$pval_file"
    CHILDREN_FILE = "$childrenmap_file"
    main(
        BASE_NAME,
        CORR_FILE,
        META_FILE,
        CMETA_FILE,
        OBSMETA_FILE,
        PVALUE_FILE,
        CHILDREN_FILE,
    )



"""
    Module that renders the Jinja templates
"""

import pathlib
from typing import Dict

from jinja2 import Environment, FileSystemLoader
from jinja2schema import infer


CONFIG_DIR: pathlib.Path = pathlib.Path(__file__).parent.parent / "config/configs"


class Template:
    """
        Base class for manipulating and rendering Jinja2 templates

        Parameters
        ----------
        template_file : pathlib.Path
            The path to the Jinja2 template file

        Attributes
        ----------
        template_vars : Set[str]
            The set of undeclared variables in the template
    """

    def __init__(self, template_file: pathlib.Path) -> None:
        loader = FileSystemLoader(str(template_file.parent))
        self._env = Environment(loader=loader)
        self._template = self._env.get_template(template_file.name)
        source, *_ = self._env.loader.get_source(self._env, template_file.name)
        self._vars = infer(source)

    def render(self, template_data: dict) -> str:
        """
            Render the template using the data passed in as arguments

            Parameters
            ----------
            template_data : dict
                Dictionary of data used to fill in the template

            Returns
            -------
            str
                The rendered template

            Raises
            ------
            UndefinedError
                If an undeclared variable is not provided a value in template_data
        """
        return self._template.render(template_data)

    @property
    def template_vars(self) -> dict:
        """
            The variables in the template
        """
        return self._vars


class ScriptTemplate(Template):
    """
        The class for templating nextflow pipeline scripts

        Parameters
        ----------
        script_file : pathlib.Path
            The path to the nextflow script template
        process_dir : pathlib.Path
            The path to the directory containing process scripts for the script_file

        Attributes
        ----------
        template_vars : Set[str]
            The set of undeclared variables in the template
        process_scripts : Dict[str, str]
            The process scripts of the nextflow pipeline
    """

    def __init__(self, script_file: pathlib.Path, process_dir: pathlib.Path) -> None:
        super().__init__(script_file)
        self._process_dir = process_dir

    @property
    def process_scripts(self) -> Dict[str, str]:
        """
            The process scripts of the nextflow pipeline
        """
        wrapper = '"""\n'
        indent = " " * 4
        scripts: Dict[str, str] = {}
        for file in self._process_dir.iterdir():
            if file.stem in self.template_vars:
                with open(file, "r") as fid:
                    line_list = fid.readlines()
                    data = "".join(map(lambda x: indent + x, line_list))
                    scripts[file.stem] = wrapper + data + indent + wrapper
        return scripts

    def render(self, template_data: dict = dict()) -> str:
        """
            Render the nextflow script template

            Parameters
            ----------
            template_data : dict, optional
                Dictionary of data used to fill in the template
                This data in addition to the process scripts that are handled automatically
                Default is an empty dictionary

            Returns
            -------
            str
                The rendered template
        """
        data = {**self.process_scripts, **template_data}
        return super().render(data)


class ResourceTemplate(Template):
    """ Class for templating nextflow resource files """

    pass


class ProfileTemplate(Template):
    """ Class for templating nextflow profile files """

    pass


class ConfigTemplate(Template):
    """
        Class for templating nextflow configuration files

        Parameters
        ----------
        config_file : pathlib.Path
            The path to the configuration file

        Attributes
        ----------
        template_vars : Set[str]
            The set of undeclared variables in the template
    """

    _resource_config: pathlib.Path = CONFIG_DIR / "resources.config"
    _profile_config: pathlib.Path = CONFIG_DIR / "profiles.config"

    def __init__(self, config_file: pathlib.Path) -> None:
        super().__init__(config_file)

    def render(self, template_data: dict, resource_config: bool = False) -> str:
        """
            Render the template using the data passed in as arguments

            Parameters
            ----------
            template_data : dict
                Dictionary of data used to fill in the template
            resource_config : bool, optional
                Flag to determine whether to append profile and resouce configs
                Default value is False

            Returns
            -------
            str
                The rendered template

            Raises
            ------
            UndefinedError
                If an undeclared variable is not provided a value in template_data
        """
        rendered_config = self._template.render(template_data)
        if resource_config:
            resource_template = ResourceTemplate(self._resource_config)
            resource = resource_template.render(template_data)
            profile_template = ProfileTemplate(self._profile_config)
            profile = profile_template.render(template_data)
        else:
            resource = ""
            profile = ""
        return rendered_config + resource + profile



#!/usr/bin/env python2

import hashlib

from biom import load_table, Table
from biom.util import biom_open
from Bio import SeqIO
from Bio.SeqIO import FastaIO


def hash_function(seq):
    return hashlib.md5(seq.encode("utf-8")).hexdigest()


def hash_otu_table(unhashed_otu_table, seqid_hash_dict, output_file):
    table = load_table(unhashed_otu_table)
    df = table.to_dataframe(dense=True)
    seq_ids = [seqid_hash_dict[i] for i in df.index]
    df.index = seq_ids
    new_table = Table(df.values, list(df.index), list(df.columns))
    with biom_open(output_file, "w") as fid:
        new_table.to_hdf5(fid, "Constructed using qiime1 clustering")


def hash_rep_seqs(unhashed_rep_seqs, output_file):
    seqs = list(SeqIO.parse(unhashed_rep_seqs, "fasta"))
    seq_ids = []
    seqid_hash_dict = dict()
    for seq in seqs:
        hash_ = hash_function(str(seq.seq))
        seqid_hash_dict[seq.id] = hash_
        seq.id = hash_function(str(seq.seq))
        seq_ids.append(seq.id)
        seq.description = ""
        seq.name = ""
    with open(str(output_file), "w") as fid:
        fasta_writer = FastaIO.FastaWriter(fid, wrap=None)
        fasta_writer.write_file(seqs)
    return seq_ids, seqid_hash_dict


def hashing(unhashed_otu_table, unhashed_rep_seqs):
    _, seqid_hash_dict = hash_rep_seqs(unhashed_rep_seqs, "rep_seqs.fasta")
    hash_otu_table(unhashed_otu_table, seqid_hash_dict, "otu_table.biom")


if __name__ == "__main__":
    UNHASHED_OTU_TABLE = "${unhashed_otu_table}"
    UNHASHED_REP_SEQS = "${unhashed_rep_seqs}"
    hashing(UNHASHED_OTU_TABLE, UNHASHED_REP_SEQS)



"""
    Console script for micone
"""

from typing import List

import click

from .logging import LOG
from .pipelines import Pipeline
from .setup import Environments
from .utils import Spinner


@click.group()
@click.option(
    "--log", "-l", default=True, type=bool, help="Flag to turn on/off logging"
)
@click.option(
    "--interactive", "-i", default=True, type=bool, help="Flag to turn on/off spinner"
)
@click.pass_context
def cli(ctx, log: bool, interactive: bool):
    """ Main entry point to micone """
    spinner = Spinner(text="Starting up...", spinner="dots", interactive=interactive)
    spinner.start()
    ctx.obj["SPINNER"] = spinner
    spinner.succeed("Successfully initialized micone")
    if log:
        LOG.enable()
        click.secho(f"Log file is at {LOG.path}")


@cli.command()
@click.option(
    "--env",
    "-e",
    default=None,
    type=str,
    help="The environment to initialize. By default will initialize all",
)
@click.pass_context
def init(ctx, env: str):
    """ Initialize the package and environments """
    spinner = ctx.obj["SPINNER"]
    environments = Environments()
    for env_cmd in environments.init(env):
        spinner.start()
        spinner.text = f"Initializing environment: {env_cmd}"
        env_cmd.wait()
        env_cmd.log()
        if env_cmd.status == "failure":
            spinner.fail(f"{env_cmd} Failed")
        elif env_cmd.status == "success":
            spinner.succeed(f"{env_cmd} Passed")
    for post_cmd in environments.post_install(env):
        spinner.start()
        spinner.text = f"Running post installation: {post_cmd}"
        post_cmd.wait()
        post_cmd.log()
        if post_cmd.status == "failure":
            spinner.fail(f"{post_cmd} Failed")
        elif post_cmd.status == "success":
            spinner.succeed(f"{post_cmd} Passed")
    click.secho(f"Log file is at {LOG.path}")
    LOG.cleanup()


@cli.command()
@click.option(
    "--profile",
    "-p",
    default="local",
    type=str,
    help="The execution profile. Either 'local' or 'sge'",
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="The config file that defines the pipeline run",
)
@click.option(
    "--output_location",
    "-o",
    type=click.Path(exists=True),
    default=None,
    help="The base output location to store pipeline results",
)
@click.option(
    "--base_dir",
    "-b",
    type=click.Path(exists=True),
    default=None,
    help="The location of base directory for input files",
)
@click.option(
    "--max_procs",
    "-m",
    type=click.INT,
    default=4,
    help="Maximum number of processes allowed to run in parallel",
)
@click.option(
    "--resume",
    is_flag=True,
    help="The flag to determine whether a previous execution is resumed",
)
@click.pass_context
def run(
    ctx,
    profile: str,
    config: click.Path,
    output_location: click.Path,
    base_dir: click.Path,
    max_procs: int,
    resume: bool,
):
    """ Run the pipeline """
    spinner = ctx.obj["SPINNER"]
    pipeline = Pipeline(
        str(config), profile, str(base_dir), resume, output_location=output_location
    )
    spinner.start()
    spinner.text = "Starting pipeline execution"
    try:
        for process in pipeline.run(max_procs=max_procs):
            if pipeline.process_queue is None:
                raise ValueError("Pipeline process queue is empty")
            process_list = " and ".join(
                [proc.id_.split(".", 2)[-1] for proc in pipeline.process_queue]
            )
            if resume and process.io_exist:
                spinner.start()
                spinner.text = f"Executing {process_list}"
                spinner.succeed(f"Resumed {process}")
            spinner.start()
            spinner.text = f"Executing {process_list}"
            updated_processes = pipeline.wait()
            for proc in updated_processes:
                proc.log()
                if proc.status == "success":
                    spinner.succeed(f"Completed {proc}")
                else:
                    spinner.fail(f"Failed to execute {proc}")
    finally:
        LOG.cleanup()


@cli.command()
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="The config file that defines the pipeline run",
)
@click.option(
    "--files", "-f", multiple=True, help="Files can be logs, configs, work, results"
)
@click.pass_context
def clean(ctx, config: click.Path, files: List[str]):
    """Clean files from a pipeline run in the current directory.
    This command irrevesibly deletes content, use with caution
    """
    click.confirm(
        f"Are you sure you want to delete {files} in current directory?", abort=True
    )
    spinner = ctx.obj["SPINNER"]
    spinner.start()
    spinner.text = "Cleaning up pipeline files"
    pipeline = Pipeline(str(config), "local")
    pipeline.clean(files)
    spinner.succeed("Completed cleanup")


def main():
    cli(obj={})


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Script that groups OTU data on different taxa levels

import json
from typing import List, Iterable, Tuple

from micone import Otu, Lineage


# Group the otu_data on all the tax_levels
def grp_otu_data(otu_data: Otu, tax_levels: List[str]) -> Iterable[Tuple[Otu, dict]]:
    sorted_tax_levels = list(reversed(tax_levels))
    child_otu = otu_data
    for tax_level in sorted_tax_levels:
        child_otu, child_groups = child_otu.collapse_taxa(tax_level)
        yield child_otu, child_groups


if __name__ == "__main__":
    TAX_LEVELS: List[str] = $tax_levels  # ['Family', 'Genus', 'Species']
    OTU_FILE = "$otu_file"  # "otu.biom"
    otu_data = Otu.load_data(OTU_FILE)
    for child_otu, child_groups in grp_otu_data(otu_data, TAX_LEVELS):
        fname = child_otu.tax_level + "_level"
        child_otu.write(fname, file_type="biom")
        with open(fname + "_children.json", "w") as fid:
            json.dump(child_groups, fid, indent=2, sort_keys=True)



"""
    Module that defines a complete pipeline by incorporating settings and processes
"""

import collections
import pathlib
import time
from itertools import chain
from typing import Deque, Dict, Iterator, List, Optional

import matplotlib.pyplot as plt
import networkx as nx
import toml

from ..config import Config
from .command import Command
from .process import Process, stringizer


class Pipeline(collections.Sequence):
    """
    Class that defines the pipeline and contains methods to run the pipeline

    Parameters
    ----------
    user_settings_file : str
        The user created settings file that describes the pipeline
    profile : {'local', 'sge'}
        The execution environment
    base_dir : str, optional
        The absolute location of the base directory for the input files
        This needs to be supplied if the input files location in the settings are relative
        If None, then current working directory is used
    resume : bool, optional
        The flag to determine whether a previous execution is resumed
        Default value is False

    Other Parameters
    ----------------
    title : str, optional
        The title of the pipeline
    project : str, optional
        The project under which to run the pipeline on the 'sge'
    order : List[str], optional
        The order of the processes in the pipeline
    output_location : str, optional
        The base output location to store all pipeline results

    Attributes
    ----------
    title : str
        The title of the pipeline
    project : str
        The project under which to run the pipeline
        Only relevant if profile is 'sge'
    output_location : str
        The base output location to store all pipeline results
    config : Config
        The configuration object for the `micone` pipelines
    profile : str
        The execution environment for the pipeline
    base_dir : pathlib.Path
        The absolute path to the base input file directory
    process_tree : nx.DiGraph
        The process tree for the pipeline
        Every node has a "process" attribute which contains the `Process` class
    """

    _req_keys = {"title", "order", "output_location"}
    process_queue: Optional[Deque[Process]] = None

    def __init__(
        self,
        user_settings_file: str,
        profile: str,
        base_dir: Optional[str] = None,
        resume: Optional[bool] = False,
        **kwargs,
    ) -> None:
        self.config = Config()
        self.profile = profile
        if self.profile == "sge":
            self._req_keys.add("project")
        self.resume = resume
        if (base_dir is None) or (base_dir == "None"):
            self.base_dir = pathlib.Path.cwd()
        else:
            base_path = pathlib.Path(base_dir)
            if base_path.is_absolute() and base_path.exists():
                self.base_dir = base_path
            else:
                raise ValueError("base_dir path must be absolute and must exist")
        user_settings = self._parse_settings(user_settings_file, **kwargs)
        title = kwargs.get("title")
        order = kwargs.get("order")
        project = kwargs.get("project")
        output_location = kwargs.get("output_location")
        self.process_tree = self._parse_process_tree(
            order if order else user_settings["order"]
        )
        self.title = title if title else user_settings["title"]
        self.project = project if project else user_settings.get("project")
        self.output_location = (
            output_location if output_location else user_settings["output_location"]
        )
        self._create_processes(user_settings)
        self._updated_processes: List[Process] = []

    @staticmethod
    def _parse_process_tree(process_string: str) -> nx.Graph:
        """
        Parses the process string and creates a process tree from it

        Parameters
        ----------
        process_string : str
            The string that represents the process tree

        Returns
        -------
        nx.Graph
            A nx.Graph representing the process tree
        """
        processes = [
            p for p in process_string.strip().replace("\n", " ").split(" ") if p
        ]
        graph = nx.DiGraph()
        root_node_name = processes[0] + ".1"
        if len(processes) == 1:
            graph.add_node(root_node_name)
            return graph
        # NOTE: We do not support forking in the first process
        process_stack = collections.deque([root_node_name])
        delimiters = {"(", ")", "|"}
        count_dict: Dict[str, int] = {}
        for process in processes[1:]:
            if process not in delimiters:
                parent = process_stack.pop()
                if process in count_dict:
                    count_dict[process] += 1
                else:
                    count_dict[process] = 1
                count = count_dict[process]
                new_process = f"{process}.{count}"
                graph.add_edge(parent, new_process)
                process_stack.append(new_process)
            elif process == "(":
                process_stack.append(process_stack[-1])
            elif process == "|":
                process_stack.pop()
                process_stack.append(process_stack[-1])
            elif process == ")":
                process_stack.pop()
        if not nx.is_directed_acyclic_graph(graph):
            raise ValueError(
                "The processes do not form a directed acyclic graph. Please check the order"
            )
        return graph

    def _parse_settings(self, settings_file: str, **kwargs) -> dict:
        """
        Parses the user created settings file

        Parameters
        ----------
        settings_file : str
            The user defined settings file that describes the pipeline

        Returns
        -------
        dict
            The dictionary of verified user settings
        """
        with open(settings_file, "r") as fid:
            settings = toml.load(fid)
        for key in self._req_keys:
            if key not in settings and key not in kwargs:
                raise ValueError(
                    f"Required key '{key}' not found in user_settings or parameters to constructor"
                )
        return settings

    def _create_processes(self, settings: dict) -> None:
        """
        Create `Process` instances and add them to the process_tree

        Parameters
        ----------
        settings : dict
            The dictionary of verified user settings
        """
        # Create processes for each node
        tree = self.process_tree
        for node_name in tree.nodes:
            suffix_flag = ""
            suffixes = [f".{i}" for i in range(1, len(tree.nodes) + 1)]
            for suffix in suffixes:
                if node_name.endswith(suffix):
                    process_name = node_name.rsplit(suffix, 1)[0]
                    suffix_flag = suffix
                    break
            else:
                process_name = node_name
            level_1, level_2, level_3, *param = process_name.split(".")
            if param:
                user_process_data = settings[level_1][level_2][level_3][param[0]]
                process_name = f"{level_1}.{level_2}.{level_3}"
            else:
                user_process_data = settings[level_1][level_2][level_3]
            default_process_data = self.config.params_set[process_name]
            default_process_data.merge(user_process_data)
            if param and suffix_flag:
                root_dir = (
                    str(default_process_data.root_dir) + f".{param[0]}" + suffix_flag
                )
            elif suffix_flag:
                root_dir = str(default_process_data.root_dir) + suffix_flag
            else:
                root_dir = str(default_process_data.root_dir)
            tree.nodes[node_name]["process"] = Process(
                default_process_data,
                self.profile,
                str(self.output_location),
                id_=node_name,
                root_dir=root_dir,
                project=self.project,
                resume=self.resume,
            )

    def __iter__(self) -> Iterator:
        return iter(self.process_tree.nodes)

    def __len__(self) -> int:
        return len(self.process_tree.nodes)

    def __getitem__(self, key: str) -> Process:
        for process in self:
            if process == key:
                return process
        raise KeyError(f"{key} is not a process of this pipeline")

    def __repr__(self) -> str:
        processes = list(self)
        return (
            f"<Pipeline title={self.title} output_location={self.output_location} "
            f"processes={processes}>"
        )

    def __str__(self) -> str:
        return self.title

    def run(self, max_procs: int = 4) -> Iterator[Process]:
        """
        Starts the execution of the pipeline
        Returns an iterator over the processes being executed

        Parameters
        ----------
        max_procs : int
            The maximum number of processes allowed to run in parallel
            Default value is 4

        Returns
        -------
        Iterator[Process]
            Iterator over each process currently being executed
        """
        # Get the process for the root node and update locations
        tree = self.process_tree
        root_node = next(nx.topological_sort(tree))
        root_node_process = tree.nodes[root_node]["process"]
        root_node_process.update_location(str(self.base_dir), "input")
        root_path = self.output_location / root_node_process.params.root_dir
        root_node_process.update_location(str(root_path), "output")
        tree.nodes[root_node]["location"] = str(root_path)
        # Attach outputs of parent node to inputs of child node
        for curr_process_name in nx.bfs_tree(tree, root_node):
            curr_process = tree.nodes[curr_process_name]["process"]
            curr_process.update_location(str(self.base_dir), "input")
            root_path = self.output_location / curr_process.params.root_dir
            curr_process.update_location(str(root_path), "output")
            tree.nodes[curr_process_name]["location"] = str(root_path)
            predecessors: List[str] = list(tree.predecessors(curr_process_name))
            while predecessors:
                prev_process_name = predecessors[0]
                prev_process = tree.nodes[prev_process_name]["process"]
                curr_process.attach_to(prev_process)
                predecessors = list(tree.predecessors(prev_process_name))
        self.draw_process_tree(self.output_location)
        self.process_queue = collections.deque([], max_procs)
        for process_name in nx.bfs_tree(tree, root_node):
            process = self.process_tree.nodes[process_name]["process"]
            loc = pathlib.Path(self.output_location)  # / process.params.output_location
            if self.resume and process.io_exist:
                yield process
            else:
                if len(self.process_queue) > self.process_queue.maxlen:
                    raise RuntimeError(
                        f"Number of parallel processes executed has exceeded {max_procs}"
                    )
                self.process_queue.append(process)
                process.build(str(loc))
                process.run()
                yield process

    def draw_process_tree(self, fpath: str) -> None:
        """
        Draw the DAG chart of the process tree

        Parameters
        ----------
        fpath: str
            The folder path where the DAG chart should be saved
        """
        tree = self.process_tree
        diagram = pathlib.Path(fpath) / "DAG.pdf"
        gml = pathlib.Path(fpath) / "DAG.gml"
        nodes = list(self.process_tree.nodes)
        labels = {n: n.split(".", 2)[-1] for n in nodes}
        if len(nodes) > 50:
            pos = nx.drawing.nx_agraph.graphviz_layout(tree, prog="twopi")
        else:
            pos = nx.drawing.nx_agraph.graphviz_layout(tree, prog="dot")
        nx.draw_networkx_nodes(tree, pos, node_size=500, alpha=0.8)
        nx.draw_networkx_edges(tree, pos, width=1.0, arrows=True)
        text = nx.draw_networkx_labels(tree, pos, labels=labels, font_size=8)
        for _, node_text in text.items():
            node_text.set_rotation(30)
        plt.axis("off")
        fig = plt.gcf()
        fig.set_size_inches(30, 20)  # default is 6.4, 4.8
        plt.savefig(diagram)
        nx.write_gml(tree, gml, stringizer=stringizer)

    @property
    def status(self) -> Dict[str, str]:
        """
        Returns the status of every process in the pipeline

        Returns
        -------
        Dict[str, str]
            A dictionary containing key=process.id_ and value=process.status
        """
        tree = self.process_tree
        root_node = next(nx.topological_sort(tree))
        status_dict: Dict[str, str] = {}
        for process_name in nx.bfs_tree(tree, root_node):
            process = tree.nodes[process_name]["process"]
            status_dict[process_name] = process.status
        return status_dict

    def wait(self, poll_rate: int = 5) -> List[Process]:
        """
        Pauses pipeline execution if process_queue is full

        Parameters
        -----------
        poll_rate : int, optional
            The numbers of seconds to sleep before polling processes for status
            Default value is 5 seconds

        Returns:
        --------
        List[Process]
            The list of processes that just finished execution
        """
        not_started_processes: List[str] = []
        self._updated_processes = []
        tree = self.process_tree
        root_node = next(nx.topological_sort(tree))
        process_order = list(nx.bfs_tree(tree, root_node))

        def _update_queue():
            running_processes: List[str] = []
            for process_id, process_status in self.status.items():
                if process_status == "in progress":
                    running_processes.append(process_id)
            for process in self.process_queue:
                if process.id_ not in running_processes:
                    self._updated_processes.append(process)
            for process in self._updated_processes:
                if process in self.process_queue:
                    self.process_queue.remove(process)

        if self.process_queue:
            if len(self.process_queue) >= self.process_queue.maxlen:
                while not self._updated_processes:
                    _update_queue()
                    time.sleep(poll_rate)
            for process_id, process_status in self.status.items():
                if process_status == "not started":
                    not_started_processes.append(process_id)
            try:
                next_process = next(
                    pid for pid in process_order if pid in not_started_processes
                )
            except StopIteration:
                while not self._updated_processes:
                    _update_queue()
                    time.sleep(poll_rate)
                return self._updated_processes
            dependent_processes = set(
                chain.from_iterable(list(tree[p.id_]) for p in self.process_queue)
            )
            while next_process in dependent_processes:
                _update_queue()
                time.sleep(poll_rate)
                dependent_processes = set(
                    chain.from_iterable(list(tree[p.id_]) for p in self.process_queue)
                )
        return self._updated_processes

    def clean(self, files: List[str] = ["logs"]) -> None:
        """
        Cleans up files from the pipeline run
        This command irrevesibly deletes content, use with caution

        Parameters
        -----------
        files : List[str]
            The supported files are:
            - logs: the `nextflow` log files
            - configs: the `nextflow` config and run scripts
            - work: the `nextflow` work directory
            - results: the results of the pipeline process
        """
        if "logs" in files:
            cwd = pathlib.Path()
            for log in chain(cwd.glob("*.log"), cwd.glob("*.log.*")):
                cmd = Command(f"rm {log}", "local")
                cmd.run()
                cmd.wait()
                cmd.log()
            cmd = Command("rm -rf .nextflow", "local")
            cmd.run()
            cmd.wait()
            cmd.log()
        if "configs" in files:
            for node in self:
                process = self.process_tree.nodes[node]["process"]
                script_file = f"{process.id_}.nf"
                config_file = f"{process.id_}.config"
                cmd = Command(f"rm {script_file} {config_file}", "local")
                cmd.run()
                cmd.wait()
                cmd.log()
        if "work" in files:
            cmd = Command("rm -rf work", "local")
            cmd.run()
            cmd.wait()
            cmd.log()
        if "results" in files:
            for node in self:
                process = self.process_tree.nodes[node]["process"]
                root_path = process.output_location / process.params.root_dir
                cmd = Command(f"rm -rf {root_path}", "local")
                cmd.run()
                cmd.wait()
                cmd.log()

    # TODO: Create computational metadata



"""
    Module to parse and store pipeline settings
"""

import collections
import os
import pathlib
import re
from subprocess import Popen, PIPE
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)


PIPELINE_DIR = pathlib.Path(__file__).parent.parent / "pipelines/src"
os.environ["PIPELINE_DIR"] = str(PIPELINE_DIR)


class Input(NamedTuple):
    """ The namedtuple class for storing input """

    datatype: str
    format: List[str]  # noqa: E701
    location: Optional[pathlib.Path] = None

    def __hash__(self) -> int:
        return hash(self.datatype)


class Output(NamedTuple):
    """ The namedtuple class for storing output """

    datatype: str
    format: List[str]  # noqa: E701
    location: pathlib.Path

    def __hash__(self) -> int:
        return hash(self.datatype)


class Parameters(NamedTuple):
    """ The namedtuple class for storing proces parameters """

    process: str
    params: dict

    def __hash__(self) -> int:
        return hash(self.process)


IOType = Union[Input, Output]


class Params(collections.Hashable):
    """
    The class for parsing and storing the parameters of a process

    Parameters
    ----------
    data : Tuple[str, Dict[str, Any]]
        The information about the parameters for a particular process

    Attributes
    ----------
    name : str
        Process name
    env : Optional[pathlib.Path]
        Location of the virtual environment for the process
    root_dir : pathlib.Path
        Directory relative to main output directory where results of the process are to be saved
    input : Set[Input]
        The list of inputs of the process
    output : Set[Output]
        The list of outputs of the process
    parameters : Set[Parameters]
        The list of parameters of the process
    """

    _req_keys = {"root_dir", "input", "output", "parameters"}

    def __init__(self, data: Tuple[str, Dict[str, Any]]) -> None:
        if len(data) != 2:
            raise ValueError(f"Invalid process data: {data}")
        key, value = data
        for req_key in self._req_keys:
            if req_key not in value:
                raise ValueError(
                    f"Invalid process data. {req_key} not found in {key} pipeline"
                )
        self.name = key
        self.root = PIPELINE_DIR / key.replace(".", "/")
        if not self.root.exists() or not self.root.is_dir():
            raise FileNotFoundError(
                f"The root directory: {self.root} doesn't exist. "
                "Please reinstall the package"
            )
        if "env" in value:
            self.env_name = value["env"]
            process = Popen(
                ["conda", "info", "-e"], shell=False, stdout=PIPE, stderr=PIPE
            )
            stdout, _ = process.communicate()
            envs = [o.strip() for o in stdout.decode("utf-8").split("\n")[2:]]
            env_loc = None
            for env in envs:
                if env.startswith(value["env"]):
                    env_loc = pathlib.Path(env.split(" ")[-1])
                    break
            self.env: Optional[pathlib.Path] = env_loc
        else:
            self.env = None
            self.env_name = "micone"
        self.root_dir = pathlib.Path(value["root_dir"])
        self.input = self._process_io(value["input"], "input")
        self.output = self._process_io(value["output"], "output")
        if not isinstance(value["parameters"], list):
            raise TypeError("Parameters must be a List.")
        self.parameters: Set[Parameters] = set()
        for curr_param in value["parameters"]:
            if "process" not in curr_param:
                raise ValueError("Parameters is missing 'process' field")
            params = {
                k: self._replace_envvar(v)
                for k, v in curr_param.items()
                if k != "process"
            }
            self.parameters.add(
                Parameters(process=curr_param["process"], params=params)
            )

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        inputs = {i.datatype for i in self.input}
        outputs = {o.datatype for o in self.output}
        return f"<Params name={self.name} input={inputs} output={outputs}>"

    def __str__(self) -> str:
        return self.name

    def get(self, name: str, category: str) -> Union[Input, Output, Parameters]:
        """
        Get Input, Output or Parameter element using its name

        Parameters
        ----------
        name : str
            The name of the IO element or parameter to be retrieved
        category: {'input', 'output', 'parameters'}
            Specifies whether the data is input, output or parameter information

        Returns
        -------
        Union[Input, Output, Parameters]
            The Input, Output or Parameter element
        """
        if category == "input":
            query_set = self.input
            attr = "datatype"
        elif category == "output":
            query_set = self.output
            attr = "datatype"
        elif category == "parameters":
            query_set = self.parameters
            attr = "process"
        else:
            raise TypeError(
                "Category can only be one of {'input', 'output', 'parameters'}"
            )
        for element in query_set:
            if getattr(element, attr) == name:
                return element
        raise KeyError(f"{name} not found in {category} of {self.name}")

    def _replace_envvar(self, item: Any) -> str:
        """
        Replace the environment variable in the "item" with it's value

        Parameters
        ----------
        item : Any
            The item which contains the environment variable
            Replacement is only done if item is a string
            Otherwise the item is returned as is

        Returns
        -------
        Any
            The item where the env variable is replaced by it's value
        """
        if isinstance(item, str):
            pattern = re.compile(r"\$\{(.*?)\}")
            if re.match(pattern, item):
                env_var = re.match(pattern, item).group(1)
                if env_var == "CONDA_PREFIX":
                    replacement = str(self.env)
                else:
                    replacement = os.environ.get(env_var)
                return re.sub(pattern, replacement, item)
            else:
                return item
        else:
            return item

    def _process_io(
        self, data: List[Dict[str, Union[str, List[str]]]], category: str
    ) -> Set[IOType]:
        """
        Process the input information

        Parameters
        ----------
        data : List[Dict[str, Union[str, List[str]]]]
            The input/output information
        category: {'input', 'output'}
            Specifies whether the data is input or output information

        Returns
        -------
        List[IOType]
            Processed input/output information
        """
        if category == "input":
            IO = Input
        elif category == "output":
            IO = Output
        else:
            raise TypeError("Category can only be either 'Input' or 'Output'")
        io_tuples: Set[IOType] = set()
        if not isinstance(data, list):
            raise TypeError(f"{category} must be a List")
        for item in data:
            req_fields = set(IO._fields) - set(IO._field_defaults.keys())
            for field in req_fields:
                if field not in item:
                    raise ValueError(f"Invalid {category}: {data}. Missing {field}")
            for field in item:
                if field not in IO._fields:
                    raise ValueError(f"Invalid {category}: {data}. Extra {field}")
            item_subset = {k: item.get(k) for k in IO._fields}
            loc: Optional[str] = item_subset.get("location")
            loc = self._replace_envvar(loc)
            item_subset["location"] = pathlib.Path(loc) if loc else None
            io_tuples.add(IO(**item_subset))
        return io_tuples

    def update_location(self, name: str, location: str, category: str) -> None:
        """
        Update the location of an Input or Output element

        Parameters
        ----------
        name : str
            The name of the IO element whose location is to be updated
        location : str
            The location of the IO element
        category: {'input', 'output'}
            Specifies whether the data is input or output information
        """
        element: Union[Input, Output] = self.get(name, category=category)
        IO: Union[Type[Input], Type[Output]]
        if category == "input":
            IO = Input
            io_list = self.input
        elif category == "output":
            IO = Output
            io_list = self.output
        else:
            raise TypeError("Category can only be either 'Input' or 'Output'")
        new_element = IO(
            datatype=element.datatype,
            format=element.format,
            location=pathlib.Path(location),
        )
        io_list.remove(element)
        io_list.add(new_element)

    def merge(self, user_settings: Dict[str, Any]) -> None:
        """
        Merge user_settings into the current Params instance

        Parameters
        ----------
        user_settings: Dict[str, Any]
            User defined settings for the current process
        """
        for curr_input in user_settings.get("input", []):
            io_item: Input = self.get(curr_input["datatype"], category="input")
            for f in curr_input["format"]:
                if f not in io_item.format:
                    raise ValueError(f"{f} not a supported format for {io_item}")
            updated_input = Input(
                datatype=io_item.datatype,
                format=curr_input["format"],
                location=pathlib.Path(curr_input["location"]),
            )
            self.input.remove(io_item)
            self.input.add(updated_input)
        for curr_params in user_settings.get("parameters", []):
            param_item: Parameters = self.get(
                curr_params["process"], category="parameters"
            )
            updated_param = Parameters(
                process=param_item.process, params={**param_item.params, **curr_params}
            )
            self.parameters.remove(param_item)
            self.parameters.add(updated_param)

    def attach_to(self, previous: "Params") -> None:
        """
        Update inputs of current `Param` instance using outputs of previous

        Parameters
        ----------
        previous : Params
            The `Param` instance to attach current instance to
        """
        for input_ in self.input:
            if not input_.location:
                try:
                    output_ = previous.get(input_.datatype, category="output")
                    self.update_location(
                        input_.datatype, str(output_.location), category="input"
                    )
                except KeyError:
                    pass


class ParamsSet(collections.Set):
    """
    The set of all supported pipeline processes

    Parameters
    ----------
    data : Dict[str, Any]
        A dictionary containing information about the pipeline processes

    Attributes
    ----------
    processes : Set[Params]
        The set of parameters for all the pipeline processes
    """

    def __init__(self, data: Dict[str, Any]) -> None:
        processed_data = {}
        for level_1 in data:
            for level_2 in data[level_1]:
                for level_3 in data[level_1][level_2]:
                    name = f"{level_1}.{level_2}.{level_3}"
                    processed_data[name] = data[level_1][level_2][level_3]
        self.processes: Set[Params] = set()
        for key, value in processed_data.items():
            process_params = Params((key, value))
            if process_params in self.processes:
                raise ValueError(
                    "Duplicate process definitions detected in settings. Aborting"
                )
            self.processes.add(process_params)

    def __iter__(self) -> Iterator:
        return iter(self.processes)

    def __len__(self) -> int:
        return len(self.processes)

    def __contains__(self, value) -> bool:
        return value in [process.name for process in self.processes]

    def __getitem__(self, key: str) -> Params:
        for process in self.processes:
            if process.name == key:
                return process
        raise KeyError(f"{key} is not in ParamsSet")

    def __repr__(self) -> str:
        processes = [process.name for process in self.processes]
        return f"<ParamsSet n={len(self)} processes={processes}>"



"""
    Module that defines different pipeline processes
"""

import collections
from copy import deepcopy
import pathlib
import re
import shutil
from typing import Any, Dict, Optional
from warnings import warn

from .command import Command
from .template import ConfigTemplate, ScriptTemplate
from ..config import Params
from ..logging import LOG


class Process(collections.Hashable):
    """
    Class for executing a pipeline process

    Parameters
    ----------
    params : Params
        The parameters for a particular process
    profile : {'local', 'sge'}
        The execution environment
    output_location : str
        The absolute path to the base input file directory
    id_ : str, optional
        The unique id_ given to the process in the process tree
        Default value is None, in this case process name is used as id_
    root_dir : str, optional
        The root directory for the results of the current process
        Default value is None
    script_name : str, optional
        The name of the process script template
        Default is 'process.nf'
    config_name : str, optional
        The name of the process configuration template
        Default is 'process.config'
    process_dir_name : str, optional
        The name of the process directory where the templates are stored
        Default is 'processes'
    project : str, optional
        The project under which to run the pipeline on the 'sge'
        Default value is None
    resume: bool, optional
        The flag to determine whether a previous execution is resumed
        Default value is False

    Attributes
    ----------
    id_ : str
        The unique id_ given to the process in the process tree
    name : str
        The name of the process
    params : Params
        The core parameters object for the process
    profile : str
        The execution environment for the process
    project : str
        The project under which to run the pipeline on the 'sge'
    script : ScriptTemplate
        The process script template
    config : ConfigTemplate
        The process configuration template
    cmd : Command
        The `Command` instance that will be executing for running the process
    env : pathlib.Path
        The location of the virtual environment
    """

    _cmd: Optional[Command] = None

    def __init__(
        self,
        params: Params,
        profile: str,
        output_location: str,
        id_: Optional[str] = None,
        root_dir: Optional[str] = None,
        script_name: str = "process.nf",
        config_name: str = "process.config",
        process_dir_name: str = "processes",
        project: Optional[str] = None,
        resume: Optional[bool] = False,
    ) -> None:
        self.params = deepcopy(params)
        self.name = self.params.name
        self.profile = profile
        self.project = project
        self.resume = resume
        self.id_ = id_ if id_ else self.name
        self.output_location = pathlib.Path(output_location)
        script_file = self.params.root / script_name
        process_dir = self.params.root / process_dir_name
        config_file = self.params.root / config_name
        self.script = ScriptTemplate(script_file, process_dir)
        self.config = ConfigTemplate(config_file)
        if root_dir:
            self.params.root_dir = pathlib.Path(root_dir)
        self.env = self.params.env

    def __hash__(self) -> int:
        return hash(self.id_)

    def __repr__(self) -> str:
        return f'<Process name={self.id_} cmd="{self.cmd}">'

    def __str__(self) -> str:
        return self.id_

    def build(self, output_dir: Optional[str] = None) -> None:
        """
        Builds the pipeline script and the config file at the output_dir
        Updates the output_location for all the results

        Parameters
        ----------
        output_dir : str, optional
            The directory where the scripts and config files are to be built
            This directory will be also be used to store the results of the process
        """
        if output_dir:
            self.output_location = pathlib.Path(output_dir)
        root_path = self.output_location / self.params.root_dir
        self.update_location(str(root_path), category="output")
        if not self.output_location.is_absolute():
            raise FileNotFoundError(f"{self.output_location} must be an absolute path")
        self.output_location.mkdir(exist_ok=True, parents=True)
        script = self.script.render()
        script_file = self.output_location / f"{self.id_}.nf"
        LOG.logger.success(f"Building script: {script_file}")
        with open(script_file, "w") as fid:
            fid.write(script)
        config = self.config.render(self.dict, resource_config=True)
        config_file = self.output_location / f"{self.id_}.config"
        LOG.logger.success(f"Building config: {config_file}")
        with open(config_file, "w") as fid:
            fid.write(config)
        work_dir = self.output_location / "work"
        if work_dir.exists():
            warning_msg = "Work directory already exists (could be from another run)"
            LOG.logger.warning(warning_msg)
        else:
            work_dir.mkdir()

    @property
    def cmd(self) -> Command:
        """
        The `Command` instance that will be used for executing the proces

        Returns
        -------
        Command
            The `Command` instance for the process
        """
        script_path = self.output_location / f"{self.id_}.nf"
        config_path = self.output_location / f"{self.id_}.config"
        log_path = self.output_location / f"{self.id_}.log"
        work_dir = self.output_location / "work"
        cmd = (
            f"nextflow -C {config_path} -log {log_path} run {script_path} -w {work_dir} "
            f"-profile {self.profile}"
        )
        # if self.resume:
        #     cmd += " -resume"
        if not self._cmd:
            self._cmd = Command(cmd, "local", timeout=100_000)
        else:
            self._cmd.update(cmd)
        return self._cmd

    def run(self) -> Command:
        """
        Starts the execution of the process and returns the Command instance
        This could produce unexpected results if run before `build`

        Returns
        -------
        Command
            The command object
            It has `cmd`, `out`, `pid` and other facilities
        """
        script_path = self.output_location / f"{self.id_}.nf"
        config_path = self.output_location / f"{self.id_}.config"
        work_dir = self.output_location / "work"
        env = self.params.env
        if not env or not env.exists() or not env.is_dir():
            raise FileNotFoundError(
                f"The environment for the process: {self.name} doesn't exist. "
                f"Please run micone init --env {self.params.env_name}"
            )
        if (
            not script_path.exists()
            or not config_path.exists()
            or not work_dir.exists()
        ):
            warning_msg = (
                "The process has not been built yet. Please run `build` before `run`"
            )
            LOG.logger.warning(warning_msg)
            warn(warning_msg)
        self.cmd.run()
        return self.cmd

    def log(self) -> None:
        """ Logs the stdout and stderr of the process to the log_file """
        LOG.logger.info(
            f"Running process: {self.id_} with profile {self.profile} and env {self.env}"
        )
        self.cmd.log()

    def wait(self) -> None:
        """ Wait for the process to complete or terminate """
        return self.cmd.wait()

    @property
    def output(self) -> str:
        """ Returns the output generated during execution of the process """
        return self.cmd.output

    @property
    def error(self) -> str:
        """ Returns the error generated during execution of the command """
        return self.cmd.error

    def clean(self, scope: str) -> None:
        """
        Cleans out the requested scope

        Parameters
        ----------
        scope : {'all', 'work_dir'}
            The scope to be cleaned
        """
        warning_msg = (
            "You are about to delete files and folders which is in irreversible process"
        )
        LOG.logger.warning(warning_msg)
        warn(warning_msg)
        if not self.output_location.is_absolute() or not self.output_location.exists():
            raise FileNotFoundError(
                f"{self.output_location} does not exist or is not an absolute path"
            )
        script_path = self.output_location / f"{self.id_}.nf"
        config_path = self.output_location / f"{self.id_}.config"
        log_path = self.output_location / f"{self.id_}.log"
        work_dir = self.output_location / "work"
        if scope == "all":
            shutil.rmtree(work_dir)
            script_path.unlink()
            config_path.unlink()
            log_path.unlink()
        elif scope == "work_dir":
            shutil.rmtree(work_dir)
        else:
            raise ValueError(
                "Unsupported scope. Please choose from {'all', 'work_dir'}"
            )

    def attach_to(self, previous: "Process") -> None:
        """
        Update inputs of current `Process` instance using outputs of previous

        Parameters
        ----------
        previous : Process
            The `Process` instance to attach the current instance to
        """
        LOG.logger.info(f"Attaching IO of {previous.name} to {self.id_}")
        self.params.attach_to(previous.params)

    def update_location(self, location: str, category: str) -> None:
        """
        Update the location of all Input and Output elements
        Useful for an intermediate process

        Parameters
        ----------
        location : str
            The location of the IO element
        category : {'input', 'output'}
            Specifies whether the data is input or output information
        """
        path = pathlib.Path(location)
        if not path.is_absolute():
            if str(path).startswith("~/"):
                str_path = str(path).replace("~/", str(path.home()))
                path = pathlib.Path(str_path)
            else:
                raise ValueError("location must be an absolute path")
        LOG.logger.info(f"Updating location of {category}s of {self.id_} to {location}")
        if category == "input":
            for input_ in self.params.input:
                in_location = input_.location
                if in_location is None:
                    continue
                if not in_location.is_absolute():
                    self.params.update_location(
                        input_.datatype, str(path / in_location), "input"
                    )
        if category == "output":
            for output_ in self.params.output:
                out_location = output_.location
                if out_location is None:
                    continue
                if not out_location.is_absolute():
                    self.params.update_location(
                        output_.datatype, str(path / out_location), "output"
                    )

    @property
    def status(self) -> str:
        """
        Return the status of the current process

        Returns
        -------
        str
            One of {'success', 'resumed', 'failure', 'in progress', 'not started'}
        """
        if self.cmd.status == "success":
            output_status = self._check_files("output")
            if output_status:
                return "success"
            return "failure"
        if self.io_exist:
            return "resumed"
        return self.cmd.status

    def _check_files(self, category: str) -> bool:
        """
        Return True if files in the category are present o/w False

        Parameters
        ----------
        category: {'input', 'output'}
            Specifies whether the data is input or output information

        Returns
        -------
        bool
            True if the expected files exist
        """
        mult_pattern = re.compile(".*{(.*)}.*")
        if category == "input":
            io_object = self.params.input
        elif category == "output":
            io_object = self.params.output
        else:
            raise ValueError(
                f"Invalid category: {category}. Can either be 'input' or 'output'"
            )
        for io_item in io_object:
            loc = io_item.location
            if loc is None:
                return False
            if "*" in str(loc):
                str_loc = str(loc)
                mult_match = re.match(mult_pattern, str_loc)
                if mult_match:
                    str_loc_list = []
                    for pattern in mult_match.group(1).split(","):
                        str_loc_list.append(re.sub(r"{.*}", pattern, str_loc))
                else:
                    str_loc_list = [str_loc]
                for str_loc in str_loc_list:
                    ind = str_loc.find("*")
                    files = list(pathlib.Path(str_loc[:ind]).glob(str_loc[ind:]))
                    if len(files) == 0:
                        return False
            elif not loc.exists():
                return False
        return True

    @property
    def io_exist(self) -> bool:
        """
        Return True if the expected input and output of the process already exist
        Useful when resuming a pipeline execution

        Returns
        -------
        bool
            True if the expected input and output of the process already exist
        """
        input_exists = self._check_files("input")
        output_exists = self._check_files("output")
        return input_exists and output_exists

    def verify_io(self) -> None:
        """
        Verify whether the Input and Output elements have been assigned and are valid
        """
        mult_pattern = re.compile(".*{(.*)}.*")
        if not self.output_location.is_absolute():
            raise ValueError("The output location must be absolute")
        for elem in self.params.input:
            if elem.location is None:
                raise ValueError(
                    f"Input: {elem} has not been assigned a location yet for process.id_: {self.id_}"
                )
            if "*" in str(elem.location):
                str_loc = str(elem.location)
                mult_match = re.match(mult_pattern, str_loc)
                if mult_match:
                    str_loc_list = []
                    for pattern in mult_match.group(1).split(","):
                        str_loc_list.append(re.sub(r"{.*}", pattern, str_loc))
                else:
                    str_loc_list = [str_loc]
                for str_loc in str_loc_list:
                    ind = str_loc.find("*")
                    files = list(pathlib.Path(str_loc[:ind]).glob(str_loc[ind:]))
                    if len(files) == 0:
                        raise FileNotFoundError(
                            f"Unable to locate input files at {elem.location} for process.id_: {self.id_}"
                        )
            elif not elem.location.exists():
                raise FileNotFoundError(
                    f"Unable to locate input file at {elem.location} for process.id_: {self.id_}"
                )
        for elem in self.params.output:
            if elem.location is None:
                raise ValueError(
                    f"Output: {elem} has not been assigned a location yet for process.id_: {self.id_}"
                )
            if not elem.location.is_absolute():
                raise ValueError(
                    "Not all the output objects have absolute paths for process.id_: {self.id_}"
                )

    @property
    def dict(self) -> Dict[str, Any]:
        """
        Return data as a dictionary

        Returns
        -------
        Dict[str, Any]
            The data stored return as a dictionary
        """
        self.verify_io()
        data: Dict[str, Any] = dict()
        data["input"] = {
            elem.datatype: str(elem.location) for elem in self.params.input
        }
        data["output"] = {
            elem.datatype: str(elem.location) for elem in self.params.output
        }
        data["output_dir"] = self.output_location / self.params.root_dir
        data["env"] = self.env
        data["project"] = self.project
        for process_params in self.params.parameters:
            data[process_params.process] = process_params.params
        return data


def stringizer(process: Process) -> str:
    """
    Convert process to string for storage in GML file

    Parameters
    ----------
    process : Process
        The process instance to be converted to string

    Returns
    -------
    str
        Stringized process
    """
    return str(process)



#!/usr/bin/env python3

from functools import partial
import os
import pathlib
import multiprocessing as mp
import shutil
import subprocess


def trim(sequence, cutadapt_args, adapters):
    front_adapter = adapters[0]
    tail_adapter = adapters[1]
    cmd = ["cutadapt"]
    if front_adapter:
        cmd.append("--front={}".format(front_adapter))
    if tail_adapter:
        cmd.append("--adapter={}".format(tail_adapter))
    cmd.extend(cutadapt_args)
    fname = sequence.rsplit("/")[-1]
    output_fname = "trimmed/" + fname
    cmd.append("--output={}".format(output_fname))
    cmd.append(sequence)
    subprocess.call(cmd)


def main(sequence_glob, trim_cmd, ncpus, adapters):
    os.mkdir("trimmed")
    shutil.copy("MANIFEST", "trimmed/MANIFEST")
    cutadapt_args = []
    with open(trim_cmd) as fid:
        for line in fid:
            cutadapt_args.append(line.strip())
    folder = pathlib.Path()
    sequences = list(str(p) for p in folder.glob(sequence_glob))
    trim_func = partial(trim, cutadapt_args=cutadapt_args, adapters=adapters)
    with mp.Pool(processes=ncpus) as pool:
        pool.map(trim_func, sequences)


if __name__ == "__main__":
    # NOTE: This template is not being used
    SEQUENCE_GLOB = "*.fastq.gz"
    TRIM_CMD = "${trim_cmd}"
    NCPUS = $ncpus
    ADAPTERS = ["${front_adapter}", "${tail_adapter}"]
    main(SEQUENCE_GLOB, TRIM_CMD, NCPUS, ADAPTERS)



#!/usr/bin/env python3

# Script that filters samples and observations based on abundance

import os

from micone import Otu


def main(otu_file, count_thres, prevalence_thres, abundance_thres):
    otu_data = Otu.load_data(otu_file)
    filtered_otu_data = otu_data.rm_sparse_obs(
        prevalence_thres, abundance_thres
    ).rm_sparse_samples(count_thres)
    fname, _ = os.path.splitext(otu_file)
    filtered_otu_data.write(f"{fname}_filtered", file_type="biom")


if __name__ == "__main__":
    OTU_FILE = "${otu_file}"
    COUNT_THRES = int("${count_thres}")
    PREVALENCE_THRES = float("${prevalence_thres}")
    ABUNDANCE_THRES = float("${abundance_thres}")
    main(OTU_FILE, COUNT_THRES, PREVALENCE_THRES, ABUNDANCE_THRES)



"""
    Module to extract, process and store pipeline settings
"""

import pathlib

import toml

from .datatypes import DataTypes
from .params import ParamsSet


CONFIG_FOLDER = pathlib.Path(__file__).parent / "configs"


class Config:
    """
        Class to extract, process and store pipeline settings

        Parameters
        ----------
        config_folder : pathlib.Path
            The path to the folder containing the configuration files

        Attributes
        ----------
        datatypes : DataTypes
            The set of supported datatypes
        params_set : ParamsSet
            The dictionary containing process configurations
    """

    def __init__(self, config_folder: pathlib.Path = CONFIG_FOLDER) -> None:
        datatypes_file = config_folder / "datatypes.toml"
        with open(datatypes_file, "r") as fid:
            datatypes = DataTypes(toml.load(fid))
        self.datatypes = datatypes
        process_types = [
            "otu_assignment",
            "tax_assignment",
            "otu_processing",
            "network_inference",
        ]
        combined_params: dict = {}
        for process_type in process_types:
            fname = config_folder / f"{process_type}.toml"
            with open(fname, "r") as fid:
                params = toml.load(fid)
            for key, value in params.items():
                combined_params[key] = value
        params_set = ParamsSet(combined_params)
        self._check_io_integrity(params_set)
        self.params_set = params_set

    def _check_io_integrity(self, params_set: ParamsSet) -> None:
        """
            Verify whether all the datatypes in the params_set are valid

            Parameters
            ----------
            params_set : ParamsSet
                The params to be verified
        """
        for param in params_set:
            for curr_input in param.input:
                if curr_input.datatype not in self.datatypes:
                    raise ValueError(
                        f"Invalid datatype {curr_input.datatype} in input definition"
                    )
                formats = self.datatypes[curr_input.datatype].format
                for curr_format in curr_input.format:
                    if curr_format not in formats:
                        raise ValueError(
                            f"Unsupported format {curr_format} in input definition"
                        )
            for curr_output in param.output:
                if curr_output.datatype not in self.datatypes:
                    raise ValueError(
                        f"Invalid datatype {curr_output.datatype} in output definition"
                    )
                formats = self.datatypes[curr_output.datatype].format
                for curr_format in curr_output.format:
                    if curr_format not in formats:
                        raise ValueError(
                            f"Unsupported format {curr_format} in output definition"
                        )
                if not curr_output.location:
                    raise ValueError(
                        f"Relative location for output {curr_output} must be defined"
                    )



"""
    Module to store pipeline datatypes
"""

import collections
from typing import Dict, Iterator, List, Set, Tuple, Union


class DataType(collections.Hashable):
    """
        The class that stores a datatype of the pipeline constructor

        Parameters
        ----------
        data : Dict[str, Any]
            The information about a particular datatype

        Attributes
        ----------
        name : str
            DataType name
        desc : str
            DataType description
        format : Set[str]
            DataType allowed formats
    """

    def __init__(self, data: Tuple[str, Dict[str, Union[str, List[str]]]]) -> None:
        if len(data) != 2:
            raise ValueError(f"Invalid DataType {data}")
        key, value = data
        if "desc" not in value or "format" not in value:
            raise ValueError(f"Invalid DataType {data}")
        self.name = key
        if isinstance(value["desc"], str):
            self.desc = value["desc"]
        else:
            raise TypeError(f"Invalid DataType {data}. Description has to be a string.")
        if isinstance(value["format"], list):
            self.format = set(value["format"])
        else:
            raise TypeError(
                f"Invalid DataType {data}. Format has to be a list of strings."
            )

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        if self.name != other.name:
            return False
        if self.desc != other.desc:
            return False
        if self.format != other.format:
            return False
        return True

    def __repr__(self) -> str:
        return f"<DataType name={self.name} desc={self.desc} format={self.format}>"

    def __str__(self) -> str:
        return self.name


class DataTypes(collections.Set):
    """
        The set of all datatypes supported by the pipeline constructor

        Parameters
        ----------
        data : Dict[str, Dict[str, Any]]
            A dictionary containing information about the pipeline datatypes
    """

    def __init__(self, data: Dict[str, Dict[str, Union[str, List[str]]]]) -> None:
        self.dtypes: Set[DataType] = set()
        for key, value in data.items():
            data_type = DataType((key, value))
            if data_type in self.dtypes:
                raise ValueError("Duplicate DataTypes detected in settings. Aborting")
            self.dtypes.add(data_type)

    def __iter__(self) -> Iterator:
        return iter(self.dtypes)

    def __len__(self) -> int:
        return len(self.dtypes)

    def __contains__(self, value: str) -> bool:
        return value in [dtype.name for dtype in self.dtypes]

    def __getitem__(self, key: str) -> DataType:
        for dtype in self.dtypes:
            if dtype.name == key:
                return dtype
        raise KeyError(f"{key} is not in DataTypes")

    def __repr__(self) -> str:
        dtypes = [dtype.name for dtype in self.dtypes]
        return f"<DataTypes n={len(self)} dtypes={dtypes}>"



#!/usr/bin/env python3

import csv
import shutil
import subprocess


def fix_manifest(manifest_file):
    data = []
    with open(manifest_file) as fid:
        header = next(fid).strip().split(",")
        for line in fid:
            if not line.startswith("#"):
                data.append(line.strip().split(","))
    header[1] = "absolute-filepath"
    for i in range(len(data)):
        fname = "demux_seqs/" + data[i][1]
        new_fname = "demux_seqs/" + data[i][0] + ".fastq.gz"
        shutil.move(fname, new_fname)
        data[i][1] = "\$PWD/" + data[i][0] + ".fastq.gz"
    with open(manifest_file, "w") as fid:
        csv_writer = csv.writer(fid, delimiter=",")
        csv_writer.writerow(header)
        csv_writer.writerows(data)


def main(demux_artifact):
    cmd = [
        "qiime",
        "tools",
        "export",
        "--input-path",
        demux_artifact,
        "--output-path",
        "demux_seqs",
    ]
    subprocess.call(cmd)
    fix_manifest("demux_seqs/MANIFEST")


if __name__ == "__main__":
    DEMUX_ARTIFACT = "${demux_artifact}"
    main(DEMUX_ARTIFACT)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['micone',
 'micone.config',
 'micone.conversion',
 'micone.logging',
 'micone.main',
 'micone.pipelines',
 'micone.pipelines.src.network_inference.bootstrap.resample.processes',
 'micone.pipelines.src.network_inference.correlation.pearson.processes',
 'micone.pipelines.src.network_inference.correlation.spearman.processes',
 'micone.pipelines.src.network_inference.direct.flashweave.processes',
 'micone.pipelines.src.network_inference.network.make_network_with_pvalue.processes',
 'micone.pipelines.src.network_inference.network.make_network_wo_pvalue.processes',
 'micone.pipelines.src.otu_assignment.chimera_checking.remove_bimera.processes',
 'micone.pipelines.src.otu_assignment.denoise_cluster.closed_reference.processes',
 'micone.pipelines.src.otu_assignment.denoise_cluster.dada2.processes',
 'micone.pipelines.src.otu_assignment.denoise_cluster.de_novo.processes',
 'micone.pipelines.src.otu_assignment.denoise_cluster.deblur.processes',
 'micone.pipelines.src.otu_assignment.denoise_cluster.open_reference.processes',
 'micone.pipelines.src.otu_assignment.sequence_processing.demultiplex_illumina.processes',
 'micone.pipelines.src.otu_assignment.sequence_processing.trim_filter_adapter.processes',
 'micone.pipelines.src.otu_assignment.sequence_processing.trim_filter_fixed.processes',
 'micone.pipelines.src.otu_processing.export.biom2tsv.processes',
 'micone.pipelines.src.otu_processing.filter.abundance.processes',
 'micone.pipelines.src.otu_processing.filter.group.processes',
 'micone.pipelines.src.otu_processing.transform.normalize.processes',
 'micone.pipelines.src.tax_assignment.assign.blast.processes',
 'micone.pipelines.src.tax_assignment.assign.naive_bayes.processes',
 'micone.pipelines.templates.denoise_cluster.chimera_checking',
 'micone.pipelines.templates.denoise_cluster.otu_assignment',
 'micone.pipelines.templates.denoise_cluster.sequence_processing',
 'micone.pipelines.templates.network_inference.bootstrap',
 'micone.pipelines.templates.network_inference.correlation',
 'micone.pipelines.templates.network_inference.direct',
 'micone.pipelines.templates.network_inference.network',
 'micone.pipelines.templates.otu_processing.export',
 'micone.pipelines.templates.otu_processing.transform',
 'micone.pipelines.templates.tax_assignment.assign',
 'micone.setup',
 'micone.utils',
 'micone.validation']

package_data = \
{'': ['*'],
 'micone.config': ['configs/*'],
 'micone.pipelines': ['configs/*',
                      'envs/micone-dada2/*',
                      'envs/micone-flashweave/*',
                      'envs/micone-mldm/*',
                      'envs/micone-netcomi/*',
                      'envs/micone-propr/*',
                      'envs/micone-qiime1/*',
                      'envs/micone-qiime2/*',
                      'envs/micone-sparcc/*',
                      'envs/micone-spieceasi/*',
                      'functions/*',
                      'modules/denoise_cluster/*',
                      'modules/denoise_cluster/chimera_checking/*',
                      'modules/denoise_cluster/otu_assignment/*',
                      'modules/denoise_cluster/sequence_processing/*',
                      'modules/network_inference/*',
                      'modules/network_inference/bootstrap/*',
                      'modules/network_inference/correlation/*',
                      'modules/network_inference/direct/*',
                      'modules/network_inference/network/*',
                      'modules/otu_processing/*',
                      'modules/otu_processing/export/*',
                      'modules/otu_processing/transform/*',
                      'modules/tax_assignment/*',
                      'modules/tax_assignment/assign/*',
                      'modules/utils/*',
                      'src/network_inference/bootstrap/pvalue/*',
                      'src/network_inference/bootstrap/pvalue/processes/*',
                      'src/network_inference/bootstrap/resample/*',
                      'src/network_inference/correlation/pearson/*',
                      'src/network_inference/correlation/propr/*',
                      'src/network_inference/correlation/propr/processes/*',
                      'src/network_inference/correlation/sparcc/*',
                      'src/network_inference/correlation/sparcc/processes/*',
                      'src/network_inference/correlation/spearman/*',
                      'src/network_inference/direct/flashweave/*',
                      'src/network_inference/direct/mldm/*',
                      'src/network_inference/direct/mldm/processes/*',
                      'src/network_inference/direct/spieceasi/*',
                      'src/network_inference/direct/spieceasi/processes/*',
                      'src/network_inference/network/make_network_with_pvalue/*',
                      'src/network_inference/network/make_network_wo_pvalue/*',
                      'src/otu_assignment/chimera_checking/remove_bimera/*',
                      'src/otu_assignment/chimera_checking/uchime/*',
                      'src/otu_assignment/chimera_checking/uchime/processes/*',
                      'src/otu_assignment/denoise_cluster/*',
                      'src/otu_assignment/denoise_cluster/closed_reference/*',
                      'src/otu_assignment/denoise_cluster/dada2/*',
                      'src/otu_assignment/denoise_cluster/de_novo/*',
                      'src/otu_assignment/denoise_cluster/deblur/*',
                      'src/otu_assignment/denoise_cluster/open_reference/*',
                      'src/otu_assignment/sequence_processing/demultiplex_454/*',
                      'src/otu_assignment/sequence_processing/demultiplex_454/processes/*',
                      'src/otu_assignment/sequence_processing/demultiplex_illumina/*',
                      'src/otu_assignment/sequence_processing/join_reads/*',
                      'src/otu_assignment/sequence_processing/join_reads/processes/*',
                      'src/otu_assignment/sequence_processing/trim_filter_adapter/*',
                      'src/otu_assignment/sequence_processing/trim_filter_fixed/*',
                      'src/otu_processing/export/biom2tsv/*',
                      'src/otu_processing/filter/abundance/*',
                      'src/otu_processing/filter/group/*',
                      'src/otu_processing/filter/partition/*',
                      'src/otu_processing/transform/normalize/*',
                      'src/tax_assignment/assign/blast/*',
                      'src/tax_assignment/assign/naive_bayes/*',
                      'src/tax_assignment/phylogeny/alignment/*',
                      'src/tax_assignment/phylogeny/diversity/*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'biom-format>=2.1.10,<3.0.0',
 'click>=7.1.2,<8.0.0',
 'ete3>=3.1.2,<4.0.0',
 'h5py>=3.1.0,<4.0.0',
 'halo>=0.0.31,<0.0.32',
 'jinja2schema>=0.1.4,<0.2.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.3.3,<4.0.0',
 'networkx>=2.5,<3.0',
 'numpy>=1.19.4,<2.0.0',
 'pandas>=1.2.0,<2.0.0',
 'schematics>=2.1.0,<3.0.0',
 'simplejson>=3.17.2,<4.0.0',
 'statsmodels>=0.12.1,<0.13.0',
 'toml>=0.10.2,<0.11.0']

extras_require = \
{'docs': ['Sphinx>=3.4.2,<4.0.0', 'sphinx-rtd-theme>=0.5.1,<0.6.0']}

entry_points = \
{'console_scripts': ['micone = micone.cli:main']}

setup_kwargs = {
    'name': 'micone',
    'version': '0.8.0',
    'description': 'The Microbial Co-occurrence Network Explorer',
    'long_description': '# MiCoNE - Microbial Co-occurrence Network Explorer\n\n![Build Status](https://github.com/segrelab/MiCoNE/workflows/build/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/micone/badge/?version=latest)](https://micone.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/segrelab/MiCoNE/branch/master/graph/badge.svg?token=2tKiI0lUJb)](https://codecov.io/gh/segrelab/MiCoNE)\n[![CodeFactor](https://www.codefactor.io/repository/github/segrelab/micone/badge)](https://www.codefactor.io/repository/github/segrelab/micone)\n[![Updates](https://pyup.io/repos/github/segrelab/MiCoNE/shield.svg)](https://pyup.io/repos/github/segrelab/MiCoNE/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n`MiCoNE`, is a flexible and modular pipeline for 16S data analysis.\nIt incorporates various popular, publicly available tools as well as custom Python modules and scripts to facilitate inference of co-occurrence networks from 16S data.\n\n<div align="center">\n⚠️ <p><strong>The package is under active development and breaking changes are possible</strong></p>\n</div>\n\n-   Free software: MIT license\n-   Documentation: <https://micone.readthedocs.io/>\n\nManuscript can be found on [bioRxiv](https://www.biorxiv.org/content/10.1101/2020.09.23.309781v2)\n\n## Features\n\n-   Plug and play architecture: allows easy additions and removal of new tools\n-   Flexible and portable: allows running the pipeline on local machine, compute cluster or the cloud with minimal configuration change. Uses the [nextflow](www.nextflow.io) under the hood\n-   Parallelization: automatic parallelization both within and across samples (needs to be enabled in the `config` file)\n-   Ease of use: available as a minimal `Python` library (without the pipeline) or the full `conda` package\n\n## Installation\n\nInstalling the minimal `Python` library:\n\n```sh\npip install micone\n```\n\nInstalling the `conda` package:\n\n```sh\ngit clone https://github.com/segrelab/MiCoNE.git\ncd MiCoNE\nconda env create -n micone -f env.yml\npip install .\n```\n\n> NOTE:\n> The `conda` package is currently being updated and will be available soon.\n\n## Workflow\n\n![pipeline](assets/pipeline.png)\n\nIt supports the conversion of raw 16S sequence data or counts matrices into co-occurrence networks through multiple methods. Each process in the pipeline supports alternate tools for performing the same task, users can use the configuration file to change these values.\n\n## Usage\n\nThe `MiCoNE` pipelines comes with an easy to use CLI. To get a list of subcommands you can type:\n\n```bash\nmicone --help\n```\n\nSupported subcommands:\n\n1. `init` - Creates `conda` environments for various pipeline processes\n2. `run` - The main subcommand that runs the pipeline\n3. `clean` - Cleans temporary data, log files and other extraneous files\n\nTo run the pipeline:\n\n```bash\nmicone run -p local -c run.toml -m 4\n```\n\nThis runs the pipeline in the `local` machine using `run.toml` for the pipeline configuration and with a maximum of 4 processes in parallel at a time.\n\n## Configuration\n\nThe configuration of the pipeline can be done using a `.toml` file.\nThe details can be found in the relevant section in the docs.\nHere is an example `config` file that performs:\n\n1. grouping of OTUs by taxonomy level\n2. correlation of the taxa using `fastspar`\n3. calculates p-values\n4. constructs the networks\n\n```toml\ntitle = "A example pipeline for testing"\n\norder = """\n  otu_processing.filter.group\n  otu_processing.export.biom2tsv\n  network_inference.bootstrap.resample\n  network_inference.correlation.sparcc\n  network_inference.bootstrap.pvalue\n  network_inference.network.make_network_with_pvalue\n"""\n\noutput_location = "/home/dileep/Documents/results/sparcc_network"\n\n[otu_processing.filter.group]\n  [[otu_processing.filter.group.input]]\n    datatype = "otu_table"\n    format = ["biom"]\n    location = "correlations/good/deblur/deblur.biom"\n  [[otu_processing.filter.group.parameters]]\n    process = "group"\n    tax_levels = "[\'Family\', \'Genus\', \'Species\']"\n\n[otu_processing.export.biom2tsv]\n\n[network_inference.bootstrap.resample]\n  [[network_inference.bootstrap.resample.parameters]]\n    process = "resample"\n    bootstraps = 10\n\n[network_inference.correlation.sparcc]\n  [[network_inference.correlation.sparcc.parameters]]\n    process = "sparcc"\n    iterations = 5\n\n[network_inference.bootstrap.pvalue]\n\n[network_inference.network.make_network_with_pvalue]\n  [[network_inference.network.make_network_with_pvalue.input]]\n    datatype = "metadata"\n    format = ["json"]\n    location = "correlations/good/deblur/deblur_metadata.json"\n  [[network_inference.network.make_network_with_pvalue.input]]\n    datatype = "computational_metadata"\n    format = ["json"]\n    location = "correlations/good/deblur/deblur_cmetadata.json"\n```\n\nOther example `config` files can be found at `tests/data/pipelines`\n\n## Credits\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter) and the [audreyr/cookiecutter-pypackage](https://github.com/audreyr/cookiecutter-pypackage) project template.\n',
    'author': 'Dileep Kishore',
    'author_email': 'dkishore@bu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/segrelab/MiCoNE',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from scipy.stats import spearmanr


def main(otu_file, output_file):
    otu_table = pd.read_table(otu_file, index_col=0)
    data = otu_table.values
    n = otu_table.shape[0]
    index = otu_table.index
    corr_data = np.zeros((n, n), dtype=np.float)
    for i in range(n):
        for j in range(i, n):
            if i == j:
                corr_data[i][j] = 1.0
            else:
                corr, _ = spearmanr(data[i, :], data[j, :])
                corr_data[i][j] = corr
                corr_data[j][i] = corr
    corr_table = pd.DataFrame(data=corr_data, index=index, columns=index)
    corr_table.fillna(value=1.0, inplace=True)
    corr_table.to_csv(output_file, sep="\\t", index=True, float_format="%.4f")


if __name__ == "__main__":
    OTU_FILE = "${otu_file}"
    OUTPUT_FILE = "${otu_file.baseName.split('_otu')[0]}_corr.tsv"
    main(OTU_FILE, OUTPUT_FILE)



#!/usr/bin/env python3

from micone import Otu


def main(
    otu_file: str,
    rm_sparse_samples: bool,
    rm_sparse_obs: bool,
    axis: str,
    count_thres: int,
    prevalence_thres: float,
    abundance_thres: float,
) -> Otu:
    otu = Otu.load_data(otu_file)
    if rm_sparse_samples:
        otu = otu.rm_sparse_samples(count_thres=count_thres)
    if rm_sparse_obs:
        otu = otu.rm_sparse_obs(
            prevalence_thres=prevalence_thres, abundance_thres=abundance_thres
        )
    return otu.normalize(axis=axis)


if __name__ == "__main__":
    OTU_FILE = "$otu_file"
    AXIS = "$axis"
    RM_SPARSE_SAMPLES = $rm_sparse_samples
    COUNT_THRES = $count_thres
    RM_SPARSE_OBS = $rm_sparse_obs
    PREVALENCE_THRES = $prevalence_thres
    ABUNDANCE_THRES = $abundance_thres
    norm_otu = main(
        OTU_FILE,
        RM_SPARSE_SAMPLES,
        RM_SPARSE_OBS,
        AXIS,
        COUNT_THRES,
        PREVALENCE_THRES,
        ABUNDANCE_THRES,
    )
    norm_otu.write("$id" + "_normalized")



#!/usr/bin/env python3

from qiime2 import Artifact


def main(input_type, input_path, input_format, output_path):
    imported_artifact = Artifact.import_data(
        input_type, input_path, view_type=input_format
    )
    imported_artifact.save(output_path)


if __name__ == "__main__":
    INPUT_TYPE = "SampleData[SequencesWithQuality]"
    INPUT_PATH = "${manifest_file}"
    INPUT_FORMAT = "SingleEndFastqManifestPhred33"
    OUTPUT_PATH = "${meta.id}_sequences.qza"
    main(INPUT_TYPE, INPUT_PATH, INPUT_FORMAT, OUTPUT_PATH)
